import string
import os
from PCAN_ISO_TP_2016 import *

IS_WINDOWS = platform.system() == 'Windows'

# Support events
if not IS_WINDOWS: 
	import select
	__LIBC_OBJ = cdll.LoadLibrary("libc.so.6")

WAIT_OBJECT_0 = 0
def NULL_HANDLE():
	return c_void_p(0) if IS_WINDOWS else c_int(0)

# Support keyboard
def getInput():
    sys.stdin.read(1)

# Isotp library
objPCANIsotp = PCAN_ISO_TP_2016()

# Help functions
def OK_KO(test):
    return "OK" if test else "KO" 

def STATUS_OK_KO(test):
    return OK_KO(objPCANIsotp.StatusIsOk_2016(test))

# Definitions

# Define Bitrate: SAE J2284-4: High-Speed CAN (HSC) for Vehicle Applications at 500 kbps with CAN FD Data at 2 Mbps 
PCAN_BITRATE_SAE_J2284_4 = create_string_buffer(b'f_clock=80000000,nom_brp=2,nom_tseg1=63,nom_tseg2=16,nom_sjw=16,data_brp=2,data_tseg1=15,data_tseg2=4,data_sjw=4')

#
# Main entry point of the program, start a small CANFD ISO TP read/write example
#

# Initialize variables 
transmitter_handle = PCANTP_HANDLE_USBBUS1 # TODO: modify the value according to your available PCAN devices.
receiver_handle = PCANTP_HANDLE_USBBUS2 # TODO: modify the value according to your available PCAN devices.

# Print version information
buffer = create_string_buffer(500)
objPCANIsotp.GetValue_2016(PCANTP_HANDLE_NONEBUS, PCANTP_PARAMETER_API_VERSION, buffer, 500)
print("PCAN-ISO-TP API Version: ", buffer.value)

# Initialize channels: CAN2.0 - 500Kbit/s
res = objPCANIsotp.InitializeFD_2016(transmitter_handle, PCAN_BITRATE_SAE_J2284_4)
print("Initialize transmitter: ", STATUS_OK_KO(res))
res = objPCANIsotp.InitializeFD_2016(receiver_handle, PCAN_BITRATE_SAE_J2284_4)
print("Initialize receiver: ", STATUS_OK_KO(res))

# Create a receive event on receiver
if IS_WINDOWS:
    receive_event = c_void_p(windll.kernel32.CreateEventA(None, 0,0,None))
    res = objPCANIsotp.SetValue_2016(receiver_handle, PCANTP_PARAMETER_RECEIVE_EVENT, receive_event, sizeof(receive_event))
    print("Create a receive event on receiver: ", STATUS_OK_KO(res))
else:
    receive_event = NULL_HANDLE()
    res = objPCANIsotp.GetValue_2016(receiver_handle, PCANTP_PARAMETER_RECEIVE_EVENT, receive_event, sizeof(receive_event))
    print("Get receive event on receiver: ", STATUS_OK_KO(res))

# Allocate tx message
tx_msg = cantp_msg()
res = objPCANIsotp.MsgDataAlloc_2016(tx_msg, PCANTP_MSGTYPE_CANFD)
print("Allocate tx message: ", STATUS_OK_KO(res))

# Allocate rx message
rx_msg = cantp_msg()
res = objPCANIsotp.MsgDataAlloc_2016(rx_msg, PCANTP_MSGTYPE_NONE)
print("Allocate rx message: ", STATUS_OK_KO(res))

# Initialize Tx message containing "42"
can_id = 0xA1
res = objPCANIsotp.MsgDataInit_2016(tx_msg, can_id, cantp_can_msgtype(PCANTP_CAN_MSGTYPE_FD.value | PCANTP_CAN_MSGTYPE_BRS.value), 1, None, None)
tx_msg.msgdata.isotp.contents.data[0] = 0x42
print("Initialize tx message: ", STATUS_OK_KO(res))

# Write "42" message
res = objPCANIsotp.Write_2016(transmitter_handle, tx_msg)
print("Write \"42\" message: ", STATUS_OK_KO(res))

# Wait a receive event on receiver
if IS_WINDOWS:
	windows_wait_result = c_uint64(windll.kernel32.WaitForSingleObject(receive_event, 100))
	wait_result = True if windows_wait_result.value == WAIT_OBJECT_0 else False
else:
	readable, _, _ = select.select([receive_event.value], [], [], 0.1)
	wait_result = True if len(readable) > 0 else False
print("Wait a message on receiver: ", OK_KO(wait_result))

# If we receive something on the receiver, read the message
if wait_result:
	res = objPCANIsotp.Read_2016(receiver_handle, rx_msg, None, PCANTP_MSGTYPE_NONE)
	print("Read message on receiver: ", STATUS_OK_KO(res))
	data = rx_msg.msgdata.any.contents.data[0]
	print("Check if the message is \"42\": ", OK_KO(data == 0x42))

# Free messages space
res = objPCANIsotp.MsgDataFree_2016(tx_msg)
print("Free tx message: ", STATUS_OK_KO(res))
res = objPCANIsotp.MsgDataFree_2016(rx_msg)
print("Free rx message: ", STATUS_OK_KO(res))

# Close receive event
if IS_WINDOWS:
	res = objPCANIsotp.SetValue_2016(receiver_handle, PCANTP_PARAMETER_RECEIVE_EVENT, NULL_HANDLE(), sizeof(NULL_HANDLE()))
	print("Stop receive event on receiver: ", STATUS_OK_KO(res))
	res_b = windll.kernel32.CloseHandle(receive_event)
	print("Close receive event: ", OK_KO(res_b))

# Uninitialize transmitter and receiver
res = objPCANIsotp.Uninitialize_2016(transmitter_handle)
print("Uninitialize transmitter: ", STATUS_OK_KO(res))
res = objPCANIsotp.Uninitialize_2016(receiver_handle)
print("Uninitialize receiver: ", STATUS_OK_KO(res))

# Exit
print("Press <Enter> to close")
getInput()
